//
// Worm epidemic model
//
// @author Michael Liljenstam, ISTS, Dartmouth College
//
// $Id: EpidemicModel.java,v 1.5 2003/02/12 22:24:19 mili Exp $

package SSF.App.Worm;

import java.io.*;
import java.util.*;

import com.renesys.raceway.SSF.*;
import SSF.Util.Random.*;
import cern.jet.random.*;

/*
import EDU.auburn.VGJ.*;
import EDU.auburn.VGJ.graph.*;
import EDU.auburn.VGJ.util.*;
import EDU.auburn.VGJ.algorithm.*;
import EDU.auburn.VGJ.algorithm.shawn.*;
import java.awt.Frame;
*/


/** 
 *  Worm epidemic model. Models the global infection spreading through 
 *  "the whole Internet" at a macroscopic level, as opposed to selective 
 *  parts where hosts and routers are modeled explicitly.
 *
 *  <p>
 *  Assumes that the worm is spreading through random scanning, and models 
 *  worm propagation through a deterministic discrete time epidemiological 
 *  model. The EpidemicModel class (singleton class) creates an internal 
 *  macroscopic representation of the Internet at the interdomain level 
 *  (AS level). The simplest representation is the "homogeneous model", 
 *  where AS boundaries are ignored and the Internet as treated as one 
 *  big network with homogeneous interactions. For someone interested in 
 *  studying interactions between ASes it's necessary to "partition" the 
 *  model up in ASes with a representation of the AS-level topology.
 *  This "stratified model" is highly experimental in the current release 
 *  code, and its use is discouraged. The ASGraph class handles loading an AS 
 *  adjacency file and building the internal topology representation.
 *  </p><p>
 *  Each AS contains an object of class WormEpidemic which handles the 
 *  epidemic state of the AS. It also contains an object of class WormTraffic 
 *  that can be used for modeling scanning traffic (although not currently 
 *  supported).
 *  </p>
 *  <p>
 *  TO DO:s <br>
 *  <ul>
 *    <li> Distributions for ASs: IP space, susceptibles, recoveries.</li>
 *    <li> "Everything is public" code clean up.</li>
 *  </ul>
 *  </p>
 */
public class EpidemicModel extends Entity {

  // model globals

  /** Time. */
  float t = 0; // in seconds

  /** Internet */
  ASGraph net;

  /** Total (initially susceptible) hosts. */
  int N;

  /** Time step. */
  float DeltaT; // set when updating

  // global epidemic state
  /** 'Overall' infection parameter. */
  public float Beta;
  /** Total susceptible hosts. */
  public float sTot;
  /** Total infected hosts. */
  public float iTot;
  /** Total recovered hosts. */
  public float rTot = 0;
  /** Cross-strata infection parameters. */
  public float[] BetaJ;

  /** Infection RNG. */
  private RandomStream infRNG;
  /** Seed for infection RNG. */
  private String       infSeed = "infSeed1";
  /** Use Binomial distr for picking number of infections. */
  Binomial             infBinomialDistr;
  /** Use a U[0,1] for picking network host. */
  Uniform              infUniformDistr;

  public HashMap asLookupFromNum;

  public float LayoutRectSize = 1000F;

  public boolean DEBUG = false;

  public boolean ssfHostsRegistered = false;

  // Classes

  /**
   *  An AS.
   */
  public class AS {

    /** Index in ASGraph array. */
    public int index;
    /** AS number. */
    public int asNumber;
    /** Adjacencies with other ASs. */ 
    public LinkedList adjacencyList;

    /** List of all susceptible hosts in this AS that occur in the network 
     *  level model. */
    public LinkedList susceptibleHostList = null;

    /** Number of infections that have been triggered in the network level 
     *  model so far. */
    public int triggeredInfections = 0;


    // layout stuff
    /** True if node has layout information stored. */
    boolean hasLayoutInfo = false;
    /** X coordinate. */
    float   x;
    /** Y coordinate. */
    float   y;

    //.............................................................
    /**
     *  Worm Epidemiological model.
     *  Contains epidemic state of the AS.
     */
    public class WormEpidemic {
      // epidemiological worm model state
      /** Susceptible hosts. */
      public float s;
      /** Infected hosts. */
      public float i;
      /** Recovered hosts. */
      public float r; 

      private AS myAS;

      public WormEpidemic(AS _myAS, int numASs) {
	myAS = _myAS;

	s = 0;
	i = 0;
	r = 0;
	    
	if (myAS.index == 0)
	  i = iTot; // all initally infected in AS 0
      }

      /** Recovery parameter function. */
      float gamma(float time) {
	  /*
	if (time > 7) 
	  return 18750F;
	else
	  */
	  return 0F;
      }

      /**
       *  One timestep.
       */
      public void update(float iTot) {

	int j = myAS.index;
	float next_s = 
	  Math.max(0, net.numASs * -BetaJ[j] * s * iTot * DeltaT + s);
	float next_i = 
	  Math.min(N, (net.numASs * BetaJ[j] * s * iTot - gamma(t)/net.numASs) * DeltaT + i); // Uniform recover distr
	float next_r = 
	  Math.min(N, gamma(t)/net.numASs * DeltaT + r);

	/*	if (DEBUG) {
	    System.out.println("t: "+t+" AS "+myAS.asNumber+" before: s="+s+", i="+i+", r="+r
		       +" - after: s="+next_s+", i="+next_i+", r="+next_r);
       } */

	// trigger infections in hosts
	myAS.triggerInfections((int) Math.floor(next_i 
						- myAS.triggeredInfections));

	// separate for clarity
	s = next_s; i = next_i; r = next_r;

      } // update


      public void output() {
	System.out.print("  " + s + "  " + i + "  " + r);
      }


      /** 
       *  Initalize parameters for worm model. Only call once. <br>
       *  Infection parameters are Uniform.<br>
       *  Susceptibles are uniformly distributed over the leaf ASs.
       */
      void init(ASGraph graph, int numASs) {

	// init infection parameters
	BetaJ  = new float[numASs];
	for (int j = 0; j < numASs; j++) {
	  BetaJ[j] = Beta / numASs; // Uniform distr of IPs
	} // for

	// init susceptibles
	LinkedList leaves = new LinkedList();
	if (numASs > 1) {
	    // stratified model
	    for (int j = 0; j < numASs; j++) {
		if (graph.cells[j].adjacencyList.size() == 1)
		    leaves.add(graph.cells[j]);
	    }
	} else {
	    // homogeneous model
	    leaves.add(graph.cells[0]);
	}
	int numLeaves = leaves.size();
	for (int i = 0; i < numLeaves; i++) {
	  AS as = (AS) leaves.removeFirst();
	  as.wormModel.s = sTot / numLeaves; // Uniform distr over leaves
	}

	if (DEBUG) {
	    System.out.println("Init worm epidemic model: s=" 
			       + sTot + ", i=" + iTot + ", r=" + rTot
			       + ", Beta=" + Beta);
	}
      } // init()
    } // class WormEpidemic

    //................................................................
    /**
     *  Worm induced traffic model.
     *  Contains information about scanning traffic induced by the worm.
     */
    public class WormTraffic {
      public AS myAS;

      // state 
      /** Mean scan rate. (Outbound) */
      public float scanRate = 0F;

      // constants
      public float sigma = 4.76F; // 4.76 probes/s per worm

      public WormTraffic(AS _myAS) {
	myAS = _myAS;
      }
      
      /** State update */
      public void update() {
	scanRate = sigma * myAS.wormModel.i; // proportional to #infected
      }
    } // class WormTraffic


    //.......AS associations and methods..................................

    // submodels
    WormEpidemic wormModel;
    WormTraffic  wormTraffic;

    public AS() {
      adjacencyList = new LinkedList();
      susceptibleHostList = new LinkedList();
    }

    public void init(int _index, int numASs) {
      index = _index;
      wormModel = new WormEpidemic(this, numASs);
      wormTraffic = new WormTraffic(this);
    }

    public void output() {
      wormModel.output();
    }

    /**
     * Signal the network model about new infections in hosts that 
     * are modeled at the network level.
     * Pick hosts at random from the susceptibleHostList in this 
     * AS.
     */
    public void triggerInfections(int numNewInfections) {
      int suscHostsInNetwModel = susceptibleHostList.size();
      if (suscHostsInNetwModel > 0) { // anything left to infect?
	int suscInAS = (int) Math.floor(wormModel.s);
	double probInfNetwModel = suscHostsInNetwModel / (double) suscInAS;
	/*     System.err.println("n " + suscHostsInNetwModel + ", p " 
	       + probInfNetwModel + " in AS " + suscInAS); */
	// number of hosts infected in the network model will have 
	// Binomial distribution
	infBinomialDistr = 
	    new Binomial(suscHostsInNetwModel, probInfNetwModel, infRNG);
	int numToInfect = infBinomialDistr.nextInt();
	
	for (int i=0; i<numToInfect; i++) {
	  int numSusceptibleHosts = susceptibleHostList.size();
	  int index = (int) Math.floor(infUniformDistr.nextDouble() 
				       * numSusceptibleHosts);
	  WormProtocolSession h = 
	      (WormProtocolSession) susceptibleHostList.remove(index);
	  h.becomeInfected();
	  triggeredInfections++;
	} // for
      } // if anything to infect
    } // triggerInfections()

    public String toString() {
      StringBuffer str = new StringBuffer();
      str.append("AS# " + asNumber + ", index " + index + " -> ");
      ListIterator it = adjacencyList.listIterator();
      while (it.hasNext()) {
	Integer as = (Integer) it.next();
	str.append(as.intValue() + ",");
      }
      return str.toString();
    }
  } // class AS ....................................................


  /**
   *  The Internet at the AS (interdomain) level.
   */
  public class ASGraph {
	
    /** Number of ASs - corresponds to: number of strata in 
	epidemiological model, number of cells in Cellular Automata. */
    public int numASs = 1;

    /** ASs */
    private AS[] cells;

    private OutputManager om = null;

    /** */
    public ASGraph(String topologyFile) {
      // init topology
      try {
	  load(topologyFile);
      } catch (IOException e) {
	  System.err.println("Couldn't load topology file: " + topologyFile);
	  System.err.println(e);
	  System.exit(1);  
      }

      // init worm model globals
      cells[0].wormModel.init(this, numASs);
    }

    public void createOutputManager(boolean stdOut, boolean globalsOut,
				    boolean doLayout, 
				    String gmlFile, String namFile, 
				    String dmlFile) {
      om = new OutputManager(this, 
			     stdOut, globalsOut, doLayout, 
			     gmlFile, namFile, dmlFile);
    }

    public void closeOutputManager() {
      om.close();
    }

    /**
     *  One timestep.
     */
    public void update() {
      iTot = 0;
      for (int j = 0; j < numASs; j++) {
	iTot += cells[j].wormModel.i;
      }

      for (int j = 0; j < numASs; j++) {
	cells[j].wormModel.update(iTot);
	cells[j].wormTraffic.update();
      }
    }


    /** Output current state. */
    void output() {
      om.output();
    }

    // state enumeration
    private static final byte AS_NUM = 0;
    private static final byte AS_OUTDEGREE = 1;
    private static final byte AS_NEIGHBOR = 2;

    /**
     *  Load topology in ASCII file adjacency list format, as downloaded from 
     *  ... (BJ?).
     */
    void load(String fileName) throws IOException {
      // if no filename is given (null), just create one big AS
      if (fileName == null) {
	  if (DEBUG) {
	      System.out.println("Homogeneous model. (Not stratfied by AS) => single AS");
	  }

	  int asNumber = 1; /// set to zero
	  AS newAS = new AS();
	  newAS.asNumber = asNumber;
	  //	  asLookupFromNum.put(new Integer(asNumber),new Integer(index)); // put into lookup table
	  numASs = 1;

	  // insert into ASGraph array and initialize
	  cells = new AS[numASs];
	  cells[0] = newAS;
	  cells[0].init(0, numASs);

	  return;
      }

      // parse input
      asLookupFromNum = new HashMap();

      System.err.println("Loading topology from file: " + fileName);

      StreamTokenizer in = new StreamTokenizer(new InputStreamReader(new FileInputStream(new File(fileName))));
      in.eolIsSignificant(true);

      int type;
      // read past first line
      while ((type = in.nextToken()) != StreamTokenizer.TT_EOF && 
	     (type = in.nextToken()) != StreamTokenizer.TT_EOL) {}

      // read each adjacency list
      byte state = AS_NUM;
      int index = 0;
      LinkedList cellList = new LinkedList();
      while ((type = in.nextToken()) != StreamTokenizer.TT_EOF) {
	switch(type) {
	case StreamTokenizer.TT_NUMBER:
	  //		System.out.print("#" + (int) in.nval + ",");
	  switch (state) {
	  case AS_NUM:
	    int asNumber = (int) in.nval;
	    if (DEBUG) {
		System.out.print("AS #" + asNumber + " connected to ");
	    }
	    //	    cells[index].asNumber = asNumber;
	    cellList.add(new AS());
	    ((AS) cellList.getLast()).asNumber = asNumber;
	    asLookupFromNum.put(new Integer(asNumber),new Integer(index)); // put into lookup table
	    state = AS_OUTDEGREE;
	    break;
  	  case AS_OUTDEGREE:
	    state = AS_NEIGHBOR;
	    break;
	  case AS_NEIGHBOR:
	    int as = (int) in.nval;
	    if (DEBUG) {
		System.out.print(as + ",");
	    }
	    //		cells[index].adjacencyList.add(new Integer(as));
	    ((AS) cellList.getLast()).adjacencyList.add(new Integer(as));
	    break;
	  }
	  break;
	case StreamTokenizer.TT_WORD:
	  System.out.println("Unexpected input from file: " + in.sval);
	  System.exit(1);
	  break;
	case StreamTokenizer.TT_EOL:
	  index++;
	  state = AS_NUM;
	  if (DEBUG)
	      System.out.println("");
	  break;
	}
      }
      if (state != AS_NUM) { // no EOL before EOF
	index++;
      }
      if (DEBUG)
	  System.out.println("");

      // Set #ASs
      numASs = index;

      // move the ASs from the cellList to the ASGraph array and initialize
      index = 0;
      cells = new AS[numASs];
      ListIterator it = cellList.listIterator();
      while (it.hasNext()) {
	cells[index] = ((AS) it.next());
	cells[index].init(index, numASs);
	index++;
      }
      System.err.println("Loaded graph of " + numASs + " ASs");
    } // load()


    public class Edge {
      public int from;
      public int to;
      public Edge(int _from, int _to) {
	from = _from; to = _to;
      }
      public String toString () {
	return ("("+from+","+to+")");
      }
    }

    public class OutputManager {
      ASGraph g;

      /** For visualization and animation files. */
      PrintWriter out;
      public boolean     stdOut = false;
      public boolean     globalsOut = false;
      public boolean     namOut = false;
      public boolean     dmlOut = false;
      WormRecorder       ssfRecorder = null;
      boolean[]          crashMarked;
      private float      globalTimeStep;
      private float      lastGlobalTimeStep;
      private float      routeUpdateBinSize = 0F;
      private int        routeUpdateBinVal = 0;
      private float      prevBinTime = 0F;
      private boolean    routingRecordOff = false;


      public OutputManager(ASGraph myASGraph,
			   boolean _stdOut, boolean _globalsOut, 
			   boolean doLayout,
			   String _gmlFile, String _namFile, 
			   String _dmlFile) {
	g = myASGraph;
	stdOut = _stdOut; globalsOut = _globalsOut;
	if (_namFile != null)
	    namOut = true;

	/*
	if (doLayout)
	    layout();
	*/

	try {
	  if (_gmlFile != null)
	    outputGML(_gmlFile + ".gml");
	  if (_dmlFile != null)
	    outputDML(_dmlFile);
	} catch (IOException e) {
	  System.err.println("ERROR: Couldn't create output file");
	  System.err.println(e);
	  System.exit(1);
	}

	if (globalsOut) {
	  // init some globals
	  globalTimeStep = DeltaT * 1000; // 1000 times the timestep
	  lastGlobalTimeStep = 0;
	}
      } // constructor

      public void close() {
      }

      /** Output current state. */
      void output() {
	if (stdOut) {
	  System.out.print(t);
	  // separate for clarity
	  for (int j = 0; j < numASs; j++) {
	    g.cells[j].output();
	  }
	  System.out.println("");
	}

	// visu output
	if (dmlOut)
	    outputDMLstate();
      }

      /**
       *  Set bin size for "global" routing update reporting, i.e. that 
       *  observed on core AS.
       */
      void setBinSize(float binSize) {
        routeUpdateBinSize = binSize;
      }

      /**
       *  Switch off output of routing update animation records.
       */
      void routingRecordOutputOff() {
	routingRecordOff = true;
      }

      /**
       *  Output graph in GML format
       */
      void outputGML(String filename) throws IOException {
	FileOutputStream os = new FileOutputStream(new File(filename));
	out = new PrintWriter(os);

	out.println("graph [");
	// output nodes
	for (int j=0; j<numASs; j++) {
	  if (!g.cells[j].hasLayoutInfo) {
	    out.println("  node [ id " + g.cells[j].index + " ]");
	  } else {
	    out.println("  node [ id " + g.cells[j].index 
			+ " graphics [ Image [ Type \"\" Location \"\" ] center [ x " + cells[j].x 
			+ " y " + cells[j].y + " z 0.0 "
			+ " ] width 10.0 height 10.0 depth 10.0 vgj [ labelPosition \"below\" shape \"Oval\" ] ] ]");
	  }
	}
	// output edges
	HashMap reverseEdges = new HashMap();
	for (int j=0; j<g.numASs; j++) {
	  ListIterator it = g.cells[j].adjacencyList.listIterator();
	  while (it.hasNext()) {
	    Integer as = (Integer) it.next();
	    int from = j;
	    int to = ((Integer) asLookupFromNum.get(as)).intValue();
	    if (reverseEdges.get((new Edge(from, to).toString())) == null) {
	      out.println("  edge [ source " + from
			  + " target " + to + " ]");
	      reverseEdges.put((new Edge(to, from)).toString(), new Integer(1));
	    }
	  }
	}

	out.println("]");

	out.close();
      } // outputGML()


      /**
       *  Output graph in DML format
       */
      void outputDML(String filename) throws IOException {
	FileOutputStream os = new FileOutputStream(new File(filename + ".dml"));
	out = new PrintWriter(os);

	float halfRect = LayoutRectSize / 2;

	out.println("Net [");
	// output nodes
	for (int j=0; j<numASs; j++) {
	  int id = cells[j].index;
	  if (!cells[j].hasLayoutInfo) {
	    out.println("  host [ id " + id 
			+ " interface [ id 0 ]"
			+ " ]");
	  } else {
	    float x = cells[j].x + halfRect;
	    float y = cells[j].y + halfRect;
	    out.println("  host [ id " + id
			+ " interface [ id 0 ]"
			+ " graphics [ x " + x
			+ " y " + y + " ] ]");
	  }
	} // for
	// output edges
	HashMap reverseEdges = new HashMap();
	for (int j=0; j<g.numASs; j++) {
	  ListIterator it = g.cells[j].adjacencyList.listIterator();
	  while (it.hasNext()) {
	    Integer as = (Integer) it.next();
	    int from = j;
	    int to = ((Integer) asLookupFromNum.get(as)).intValue();
	    if (reverseEdges.get((new Edge(from, to).toString())) == null) {
	      out.println("  link [ attach " + from
			  + "(0) attach " + to + "(0) ]");
	      reverseEdges.put((new Edge(to, from)).toString(), new Integer(1));
	    }
	  }
	}

	out.println("]");

	out.close();

	// init visualization state
	dmlOut = true; // set DML output flag
	ssfRecorder = new WormRecorder(filename);
	crashMarked = new boolean[numASs];
	for (int i=0; i<numASs; i++) {
	  crashMarked[i] = false;
	}
      } // outputDML()


      /**
       *  SSFNet visualization output.
       */
      void outputDMLstate() {
	// global state indicators
	if (globalsOut && t >= (lastGlobalTimeStep + globalTimeStep)) {
	  // infections
	  float totalInfectedHosts = 0;
	  for (int j=0; j<g.numASs; j++) {
	    totalInfectedHosts += g.cells[j].wormModel.i;
	  }
	  ssfRecorder.
	      sendGlobalRecord(t, totalInfectedHosts);

	  lastGlobalTimeStep = t;
	} // if
	if (globalsOut) {
	  // specialized - pick an AS from the "core" and observe 
	  // routing traffic
	  int observeAS = -1;
	  switch (numASs) {
	  case 68: // "99%" graph
	    observeAS = 44; // index
	    break;
	  case 194: // "94%" graph
	    observeAS = 18; // index
	    break;
	  case 799: // "75%" graph
	    observeAS = 221; // index
	    break;
	  default:
	  }
	  /*
	  if (observeAS >= 0) {
	    int activity = 
		g.cells[observeAS].gwRoutingState.routeUpdates.size();
	    if (routeUpdateBinSize > 0) {
	      if (t < (prevBinTime + routeUpdateBinSize)) {
		routeUpdateBinVal += activity;
	      } else {
		if (routeUpdateBinVal > 0) {
		    ssfRecorder.
			sendGlobalRoutingTrafficRecord(t, observeAS, routeUpdateBinVal);
		}
		prevBinTime = t;
		routeUpdateBinVal = 0;
	      } 
	    } else {
	      if (activity > 0) {
		ssfRecorder.
		  sendGlobalRoutingTrafficRecord(t, observeAS, activity);
	      } // if activity
	    }
	  }
	  */
	}
      } // outputDMLstate()


	/*
      public class VGJ_Graph {
	public Graph g;

	public VGJ_Graph(ASGraph asGraph) {
	  g = new Graph();

	  // insert nodes
	  for (int j=0; j<asGraph.numASs; j++) {
	    g.insertNode();
	  }
	  // insert edges
	  HashMap reverseEdges = new HashMap();
	  for (int j=0; j<asGraph.numASs; j++) {
	    ListIterator it = asGraph.cells[j].adjacencyList.listIterator();
	    while (it.hasNext()) {
	      Integer as = (Integer) it.next();
	      int from = j;
	      int to = ((Integer) asLookupFromNum.get(as)).intValue();
	      if (reverseEdges.get((new Edge(from, to).toString())) == null) {
		g.insertEdge(from, to);
		reverseEdges.put((new Edge(to, from)).toString(), new Integer(1));
	      }
	    } // while
	  } // for
	} // constructor
      } // VGJ_Graph


      public class DummyUpdate implements GraphUpdate {
	public void update(boolean adjust_bounds) { System.err.print("."); }
	public void scale(double scaleval) {}
	public void center() {}
	public DRect windowRect() 
	  { return new DRect(0,0,LayoutRectSize,LayoutRectSize); }
	public double getHSpacing() {return 0.0;}
	public double getVSpacing() {return 0.0;}
	public Frame getFrame() { return new Frame(); }
	public Node getSelectedNode() { return new Node(); }
      }
	*/

      /**
       *  Do a layout of the graph using the VGJ Spring algorithm.
       */
	/*
      void layout() {
	VGJ_Graph vgj = new VGJ_Graph(g);

	// layout graph
	DummyUpdate dummy = new DummyUpdate();
	Spring s = new Spring();
	System.err.println("Doing graph layout...");
	//      s.compute(g, dummy);
	s.compute(vgj.g, dummy);
	System.err.println("");

	// update AS nodes
	for (int j=0; j<g.numASs; j++) {
	  //	DPoint p = g.getNodeFromIndex(j).getPosition();
	  DPoint p = vgj.g.getNodeFromIndex(j).getPosition();
	  cells[j].x = (float) p.x;
	  cells[j].y = (float) p.y;
	  cells[j].hasLayoutInfo = true;
	}
      } // layout()
	*/
    } // OutputManager

  } // ASGraph

  //

  EpidemicModel(String topologyFile, String rngSeed, boolean debugOn, 
		int s_0, int i_0, float _beta) {
    DEBUG = debugOn;
    if (DEBUG) {
	System.out.println("Creating epidemic model from topology file " 
			   + topologyFile);
    }

    // init RNG
    if (rngSeed != null) {
	infSeed = rngSeed;
    }
    try {
      infRNG = RandomStream.getStream(this,
				      "MersenneTwister",
				      infSeed);

    } catch (RandomStream.StreamException rngex) {
      rngex.printStackTrace();
      System.exit(-1);
    }
    infUniformDistr = new Uniform(0, 1, infRNG);

    // startup values for epidemic model
    sTot = s_0; 
    iTot = i_0;
    Beta = _beta;
    N = s_0 + i_0 + (int) Math.round(rTot); // total number of hosts in system

    // init network
    net = new ASGraph(topologyFile);
  }


  //
  // Methods for programmatic use
  //
    
  /**
   *  Register a host, actually its WormProtocolSession, with the global 
   *  epidemic model.
   *  @param _host  The calling host.
   */
  public void registerHost(WormProtocolSession _host, int asn) {
    if (DEBUG) {
      String hostNH = ((SSF.Net.Host) _host.inGraph()).nhi;
      System.err.println("Registering host " + hostNH + " in ASN " + asn);
    }
    int index;
    if (net.numASs == 1) { 
	// homogeneous model
	index = asn-1; // if non-stratified model
	net.cells[index].asNumber = asn;
    } else {
	// stratified model - lookup index from lookup table
	index = ((Integer) asLookupFromNum.get(new Integer(asn))).intValue();
	// asNumber attr already set
    }
    net.cells[index].susceptibleHostList.add(_host);
    ssfHostsRegistered = true;
  }

  /**
   *  Update the timestepped model.
   *  @param newTime  Step forward one step to this time (in hours).
   *  @return New simulation time.
   */
  public float update(float newTime) {
    DeltaT = newTime - t;
    net.update();
    t = newTime;
    return t;
  }

  /**
   *  Get the total number of infectioned hosts in the whole system.
   */
  public float getTotalInfections() {
    float iTot = 0;
    for (int j = 0; j < net.numASs; j++) {
      iTot += net.cells[j].wormModel.i;
    }
    return iTot;
  }
}
