//
// WormProtocolSession.java
// A pseudo-protocol that models a worm.
// 
//
// @author Michael Liljenstam, ISTS, Dartmouth College
//
// $Id: WormProtocolSession.java,v 1.3 2003/02/12 21:38:46 mili Exp $

package SSF.App.Worm;

import SSF.OS.*;
import SSF.Net.*;
import SSF.Net.Util.*;
import com.renesys.raceway.DML.*;
//import java.util.*;

//import SSF.OS.BGP4.*;
import SSF.OS.BGP4.Util.*;

// ==== SSF.App.Worm.WormProtocolSession class ==========================/
/**
 * Models the worm infestation in a single host. Connects to the global 
 * epidemic model to receive infection trigger. See <tt>config()</tt> method 
 * for more information on how to set the parameters for the epidemic model.
 *
 * @author Michael Liljenstam, ISTS, Dartmouth College
 */
public class WormProtocolSession extends ProtocolSession 
    implements Configurable {

  /************************** Global model ******************************/

  /** Reference to the global epidemic model. */
  static EpidemicModel        wormModel;

  /** Switch to select if a homogeneous epidemic model should be used 
   *  (default), or a model that is stratified by AS. */
  static boolean              stratifiedOn = false;

  /** Name of AS topology file, if stratified model is used. */
  static String               asGraphFileName = null;

  /** Seed string for infection Random Number stream. */
  static String               infectionRNGseed = null;

  // ....... start values for epidemic model........
  /** Initial S, i.e. initial number of susceptible hosts. */
  static int                  s_0; // required param
  /** Initial I, i.e. initial number of infected hosts. */
  static int                  i_0 = 1; // default = 1 infected host
  /** Beta, the infection parameter. Controls infection rate. */
  static float                beta = 0; // required param

  //...... Enumeration of states ......................................../
  static final byte   SUSCEPTIBLE = 0;
  static final byte   INFECTED    = 1;
  static final byte   REMOVED     = 2;


  /************************** Attributes ********************************/

  /** Host entity where this pseudo-protocol is installed */
  Host               localHost;

  /** Host's NHI address */
  String             localNHI;

  /** Time step. */
  float              deltaT = 60.0F; // 1 min

  /** AS number this router belongs to. */
  int                asn;

  /** Print out debug information. */
  boolean            debug = false;

  /** Epidemiological state. */
  byte               state = SUSCEPTIBLE;

  // .......... Helper classes ........................................../

  /**
   * Timer that drives the global epidemic model.
   */
  public class WormProtocolSessionTimer extends Timer {
    ProtocolSession mySess = null;
      
    public WormProtocolSessionTimer(ProtocolGraph e, long dt, 
				    ProtocolSession sess) {
      super(e, dt);
      mySess = sess;
    }

    public void callback() {
      ((WormProtocolSession) mySess).update();
      set();
    }
  }

  /************************** Methods    ********************************/

  public WormProtocolSession() {
    super();
    System.err.println("WormProtocolSession created");
  }



  /** 
   *  Initalize (after config phase). First instance creates global epidemic 
   *  model object and initiates the timer that drives the epidemic model. 
   *  All instances register with the epidemic model so that it can find them 
   *  to trigger infection events.
   */
  public void init() {
    // create worm model
    if (wormModel == null) {
      // create worm model
      wormModel = new EpidemicModel(asGraphFileName, infectionRNGseed, debug, 
				    s_0, i_0, beta);

      // setup "driver" timer
      (new WormProtocolSessionTimer(inGraph(), SSF.Net.Net.seconds(deltaT), 
				    this))
	  .set(SSF.Net.Net.seconds(deltaT));

      // setup "init" timer
      /*
      (new Timer(inGraph(), 0) {
	    public void callback() {
		WormProtocolSession.wormModel.initInfect();
	    }
	  }).set();
      */
    }

    // register this router with the worm model
    if (!stratifiedOn) {
	wormModel.registerHost(this, 1); // set ASN=1 for all
    } else {
	// find out ASN
	asn = AS_descriptor.nh2as(AS_descriptor.get_as_nh((Host) inGraph()));
	// register
	wormModel.registerHost(this, asn);
    }
  }


  /** 
   *  Configure WormProtocolSession. Also sets parameters for the global 
   *  epidemic.
   *  <p>
   *  REQUIRED ATTRIBUTES: <br>
   *  <ul>
   *    <li><b><i>s_0</i></b>    --- inital number of suscepible hosts in 
   *                                  the whole of the Internet.</li>
   *    <li><b><i>i_0</i></b>    --- inital number of infected hosts</li>
   *    <li><b><i>beta</i></b>   --- infection parameter.</li>
   *
   *  </ul><p>
   *
   *  OPTIONAL ATTRIBUTES:<br> 
   *  <ul>
   *    <li><b>debug</b>         --- "debug true" will switch on debugging 
   *                                 printouts for this session and the 
   *                                 global epidemic process.</li>
   *    <li><b>delta_t</b>       --- set the time step size (in seconds).</li>
   *    <li><b>infection_seed</b> --- seed string for the infection process 
   *                                  in the epidemic model.</li>
   *    <li><b>stratified_on</b> --- stratified model (experimental).</li>
   *    <li><b>as_graph</b>      --- name of AS graph file (connectivity).
   *                                 Required for stratified model.</li>
   *
   *  </ul><p>
   */
  public void config(Configuration cfg) throws configException {
    super.config(cfg);

    //    System.err.println("WormProtocolSession config");

    localHost = (SSF.Net.Host) inGraph();
    localNHI = localHost.nhi;

    // Infection RNG seed string?
    try {
      infectionRNGseed = (String) cfg.findSingle("infection_seed");
    } catch (configException e) {
	System.err.println("WormProtocolSession: " + e);
    }

    // time step
    try {
      if (cfg.findSingle("delta_t") != null) {
        deltaT = Float.parseFloat((String)cfg.findSingle("delta_t"));
	System.err.println("WormProtocolSession deltaT: " + deltaT);
      } 
    } catch (configException e) {
	System.err.println("WormProtocolSession: " + e);
    }

    // init epidemic model parameters
    try {
      String s0_str = (String) cfg.findSingle("s_0");
      if(s0_str != null) {
	  s_0 = Integer.parseInt(s0_str);
      } else {
	  System.err.println("Config error in WormProtocolSession: " 
			     + "'s_0' is a required attribute");
	  System.exit(-1);
      }
    } catch (configException e) {
	System.err.println("WormProtocolSession: " + e);
    }
    try {
      String i0_str = (String) cfg.findSingle("i_0");
      if(i0_str != null) {
	  i_0 = Integer.parseInt(i0_str);
      }
    } catch (configException e) {
	System.err.println("WormProtocolSession: " + e);
    }
    try {
      String beta_str = (String) cfg.findSingle("beta");
      if(beta_str != null) {
	  beta = Float.parseFloat(beta_str);
      } else {
	  System.err.println("Config error in WormProtocolSession: " 
			     + "'beta' is a required attribute");
	  System.exit(-1);
      }
    } catch (configException e) {
	System.err.println("WormProtocolSession: " + e);
    }

    // look for debug flag
    try {
      if((cfg.findSingle("debug") != null) && 
	 (((String)cfg.findSingle("debug")).compareTo("true") == 0))
	  debug =  true;
    } catch (configException e) {
	System.err.println("WormProtocolSession: " + e);
    }

    // stratify model by AS? default is homogeneous (non-stratified)
    try {
      if((cfg.findSingle("stratified_on") != null) && 
	 (((String)cfg.findSingle("stratified_on")).compareTo("true") == 0))
	  stratifiedOn =  true;
    } catch (configException e) {
	System.err.println("WormProtocolSession: " + e);
    }

    // set AS graph file (required if stratified model)
    try {
      asGraphFileName = (String) cfg.findSingle("as_graph");
      if(stratifiedOn && asGraphFileName == null) {
	  System.err.println("Config error in WormProtocolSession " 
			     + localNHI 
			     + ": 'as_graph' is a required attribute for stratified model");
	  System.exit(-1);
      }
    } catch (configException e) {
	System.err.println("WormProtocolSession: " + e);
    }

  } // config

  /**
   *  Do nothing.
   */
  public boolean push(ProtocolMessage msg, ProtocolSession fromSessn) {
    return true;
  }

  /**
   * getSimTime: get the simulation time
   * it's the same function used in tcpSessionMaster.java
   */
  public float getSimTime() {
    return inGraph().now()/(float)SSF.Net.Net.seconds(1);
  }

  /**
   * Do one timestep update of the worm model.
   * 
   */
  public void update() {
    float t = getSimTime(); // time in secs
    float newWormModelTime = wormModel.update(t);
    if (debug)
      System.err.println("t: " + getSimTime() 
			 + " Worm protocol session update, new t=" 
			 + newWormModelTime
			 + " s, i=" + wormModel.getTotalInfections());
  }


  /**
   * Make this host become infected.
   * 
   */
  public void becomeInfected() {
    switch (state) {
    case SUSCEPTIBLE:
      if (debug)
	System.err.println("t: " + getSimTime() 
			   + " Host " + ((Host) this.inGraph()).nhi 
			   + " infected");
      state = INFECTED;
      //      wormModel.registerInfection(this);
      break;
    default:
      System.err.println("Unexpected state: " + state);
      System.exit(-1);
    }
  }


  /**
   *  Get epidemiological state.
   * 
   */
  public byte getState() {
    return state;
  }
}
